/*
 * iSCSI driver for Linux
 * Copyright (C) 2002 Cisco Systems, Inc.
 * maintained by linux-iscsi-devel@lists.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * $Id: iscsi-linux.c,v 1.1.2.9 2004/09/14 21:18:02 mikenc Exp $
 *
 */

#include <stdarg.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/utsname.h>

#include "iscsi-ioctl.h"
#include "iscsi-version.h"
#include "iscsi-hooks.h"
#include "iscsi-login.h"

const char *progname = "iscsid";

/* set by signal handlers */
static int stop_requested = 0;
static int reload_config = 0;

char *build_str = BUILD_STR;

#define ISCSI_DEVICE "/dev/iscsictl"

static int control_fd = -1;

void
sigusr1_handler(int unused)
{
	daemon_config.debug_level++;
	debugmsg(1, "received sigusr1, debug level now %d",
		 daemon_config.debug_level);
}

void
sigusr2_handler(int unused)
{
	if (daemon_config.debug_level > 0)
		daemon_config.debug_level--;
	debugmsg(1, "received sigusr2, debug level now %d",
		 daemon_config.debug_level);
}

void
sigterm_handler(int unused)
{
	debugmsg(1, "received sigterm", getpid());
	stop_requested = 1;
}

void
sighup_handler(int unused)
{
	debugmsg(1, "received sighup", getpid());
	reload_config = 1;
}

/* Note: test and clear semantics */
int
iscsi_should_reload_config(void)
{
	int ret = reload_config;

	reload_config = 0;

#ifdef RESET_LUN_PROBING
	/* see if it works better to have the kernel module automatically reset
	 * whenever the queue empties, and have it ensure the same session is
	 * never queued more than once. It's easier to track this in the kernel,
	 * and we avoid hanging the daemon if a kernel module bug prevents this
	 * reset from ever succeeding.
	 */
	if (ret) {
		/* reset LUN probing now, so that the re-probes we do after
		 * reloading the config occur in a reasonable order.
		 */
		while (ioctl(control_fd, ISCSI_RESET_PROBING, NULL) <= 0) {
			if (stop_requested)
				exit(0);

			usleep(10 * 1000);
		}
		debugmsg(1, "reset LUN probing before reloading config");
	}
#endif

	return ret;
}

int
iscsi_process_should_exit(void)
{
	return stop_requested;
}

int
open_control_device(void)
{
	int ctlfd;

	ctlfd = open(ISCSI_DEVICE, O_RDWR);
	if (ctlfd < 0) {
		errormsg("cannot open %s", ISCSI_DEVICE);
		return -1;
	}

	return ctlfd;
}

/* initialize the daemon_config to it's default state,
 * process the command line, and do any other initialization needed
 * for the main daemon process.
 */
void
iscsi_daemon_starting(int argc, char **argv)
{
	int c;
	struct sigaction action;
	struct utsname host_info;
	int retry_count;

	/* signal handling */
	memset(&action, 0x0, sizeof (action));
	action.sa_sigaction = NULL;
	action.sa_flags = 0;

	action.sa_handler = sigusr1_handler;
	sigaction(SIGUSR1, &action, 0);

	action.sa_handler = sigusr2_handler;
	sigaction(SIGUSR2, &action, 0);

	action.sa_handler = sighup_handler;
	sigaction(SIGHUP, &action, 0);

	action.sa_handler = sigterm_handler;
	sigaction(SIGTERM, &action, 0);
	sigaction(SIGINT, &action, 0);

	openlog(progname, LOG_PID, LOG_DAEMON);

	/* configure default files */
	daemon_config.pid_file = "/var/run/iscsid.pid";
	daemon_config.config_file = "/etc/iscsi.conf";
	daemon_config.initiator_name_file = "/etc/initiatorname.iscsi";
	/* dup it so that the memory is writable */

	/* determine InitiatorName or exit */
	daemon_config.initiator_name =
	    get_iscsi_initiatorname(daemon_config.initiator_name_file);
	if (daemon_config.initiator_name == NULL) {
		logmsg(AS_NOTICE, "daemon exiting due to configuration error");
		exit(1);
	}

	/* optional InitiatorAlias */
	memset(&host_info, 0, sizeof (host_info));
	if (uname(&host_info) >= 0) {
		daemon_config.initiator_alias = strdup(host_info.nodename);
	}

	daemon_config.debug_level = 0;

	/* process any command line arguments */
	while ((c = getopt(argc, argv, "d:nvf:s:")) >= 0) {
		switch (c) {
		case 'd':
			if (optarg) {
				daemon_config.debug_level = atoi(optarg);
			} else
				daemon_config.debug_level++;
			debugmsg(1, "iSCSI debug level %d",
				 daemon_config.debug_level);
			break;
		case 'f':
			if (optarg) {
				daemon_config.config_file = strdup(optarg);
				debugmsg(1, "using configuration file %s",
					 daemon_config.config_file);
			}
			break;
		case 'n':
			daemon_config.foreground = 1;
			debugmsg(1, "daemon will stay in foreground");
			break;
		case 'v':
			printf("iscsid version %s (%s)\n",
			       ISCSI_DRIVER_VERSION, ISCSI_DATE);
			if (build_str)
				/* developer-built internal release */
				printf("%s\n", build_str);
			exit(0);
		default:
			logmsg(LOG_ERR,
			       "iscsid - unexpected option %d, '%c', exiting",
			       c, c);
			exit(1);
		}
	}

	debugmsg(1, "InitiatorName=%s", daemon_config.initiator_name);
	debugmsg(1, "InitiatorAlias=%s", daemon_config.initiator_alias);

	/* log the version, so that we can tell if the daemon and kernel module 
	 * match based on what shows up in the syslog.  Tarballs releases 
	 * always install both, but Linux distributors may put the kernel module
	 * in a different RPM from the daemon and utils, and users may try to 
	 * mix and match in ways that don't work.
	 */
	logmsg(LOG_NOTICE, "version %s variant (%s)",
	       ISCSI_DRIVER_VERSION, ISCSI_DATE);
	if (build_str)
		logmsg(LOG_NOTICE, build_str);

	/* ensure the control device exists, and open it for all of the 
	 * child processes to use 
	 */
	retry_count=0;
	for (;;) {
		control_fd = open_control_device();
		if (control_fd >= 0)
			break;
		logmsg(AS_ERROR,
		       "could not open control device %s.  "
		       "Make sure iscsi is loaded",
		       ISCSI_DEVICE);
		if (stop_requested)
			exit(0);
		sleep(10);
		retry_count++;
		if (retry_count > 2) { /* 20 seconds max */
			exit(0);
		}
	}

	retry_count=0;

#ifdef RESET_LUN_PROBING
	/* reset the LUN probing, to handle cases where the daemon restarts 
	 * without reloading the kernel module 
	 */
	retry_count=0;
	while (ioctl(control_fd, ISCSI_RESET_PROBING, NULL) <= 0) {
		debugmsg(1,
			 "failed to reset LUN probing while daemon starting\n");

		if (stop_requested)
			exit(0);

		usleep(10 * 1000);
		retry_count++;
		if (retry_count > 50) { /* 500ms max */
			exit(0);
		}
	}
	debugmsg(1, "reset LUN probing while daemon starting");
#endif

}

void
iscsi_daemon_stopping(void)
{
	close(control_fd);
	/* nothing to do at the moment */
}

void
iscsi_init_config_defaults(struct iscsi_config_defaults *defaults)
{
	/* enabled/disabled */
	defaults->enabled = 1;

	/* discovery defaults */
	defaults->continuous_sendtargets = 1;	/* auto detect */
	defaults->send_async_text = 1;
#ifdef SLP_ENABLE
	defaults->slp_multicast = 1;
#else
	defaults->slp_multicast = 0;
#endif
	defaults->slp_scopes = NULL;
	defaults->slp_poll_interval = 5 * 60;	/* 5 minutes */

	/* auth options */
	defaults->auth_options.authmethod = CHAP_AUTHENTICATION;
	defaults->auth_options.password_length = 0;
	defaults->auth_options.password_length_in = 0;

	/* connection timeouts */
	defaults->connection_timeout_options.login_timeout = 15;
	defaults->connection_timeout_options.auth_timeout = 45;
	defaults->connection_timeout_options.active_timeout = 5;
	defaults->connection_timeout_options.idle_timeout = 60;
	defaults->connection_timeout_options.ping_timeout = 5;

	/* error timeouts */
	defaults->error_timeout_options.abort_timeout = 10;
	defaults->error_timeout_options.reset_timeout = 30;

	/* session timeouts */
	defaults->session_timeout_options.replacement_timeout = 0;	
		/* infinite retries (no timeout) by default, for safety */
	defaults->session_timeout_options.disk_command_timeout = 0;	
		/* no maximum enforced */

	/* tcp options */
	defaults->tcp_options.window_size = 256 * 1024;

	/* iSCSI operational parameters */
	defaults->iscsi_options.InitialR2T = 0;
	defaults->iscsi_options.ImmediateData = 1;
	defaults->iscsi_options.MaxRecvDataSegmentLength = 128 * 1024;
	defaults->iscsi_options.FirstBurstLength = 256 * 1024;
	defaults->iscsi_options.MaxBurstLength = (16 * 1024 * 1024) - 1024;
	defaults->iscsi_options.DefaultTime2Wait = 0;	
				/* we only use session reinstatement (ERL 0) */
	defaults->iscsi_options.DefaultTime2Retain = 0;	
				/* we only use session reinstatement (ERL 0) */
	defaults->iscsi_options.HeaderDigest = CONFIG_DIGEST_PREFER_OFF;
	defaults->iscsi_options.DataDigest = CONFIG_DIGEST_PREFER_OFF;

	/* multipath options */
	defaults->multipath_options.type = NO_MULTIPATH;

	/* Portal Failover */
	defaults->portal_failover = 1;

	/* Preferred portal/Subnet */
	defaults->preferred_portal = NULL;
	defaults->preferred_subnet = NULL;
	defaults->preferred_subnet_mask = 0;
}

static int
ioctl_establish_session(struct iscsi_session_config *config, int probe_luns,
			int config_number, int update)
{
	int rc, ret = 0;
	int p = 0;
	struct iscsi_portal_config *portal = config->portals.head;
	struct iscsi_session_ioctl *ioctld = NULL;
	struct sockaddr_in *addr;

	p = 0;
	for (portal = config->portals.head; portal; portal = portal->next) {
		p++;
	}

	/* allocate an ioctl structure with enough space for 
	 * all of the portal info 
	 */
	ioctld =
	    calloc(1,
		   sizeof (*ioctld) + (p * sizeof (struct iscsi_portal_info)));
	if (ioctld == NULL) {
		logmsg(AS_ERROR,
		       "failed to allocate %d bytes for ioctl data structure",
		       sizeof (*ioctld) +
		       (p * sizeof (struct iscsi_portal_info)));
		return 1;
	}

	ioctld->ioctl_size = sizeof (*ioctld);
	ioctld->ioctl_version = ISCSI_SESSION_IOCTL_VERSION;
	ioctld->config_number = config_number;
	ioctld->update = update;
	ioctld->portal_info_size = sizeof (struct iscsi_portal_info);
	ioctld->num_portals = p;

	p = 0;
	for (portal = config->portals.head; portal; portal = portal->next, p++) {
		ioctld->portals[p].login_timeout =
		    portal->connection_timeout_options.login_timeout;
		ioctld->portals[p].auth_timeout =
		    portal->connection_timeout_options.auth_timeout;
		ioctld->portals[p].active_timeout =
		    portal->connection_timeout_options.active_timeout;
		ioctld->portals[p].idle_timeout =
		    portal->connection_timeout_options.idle_timeout;
		ioctld->portals[p].ping_timeout =
		    portal->connection_timeout_options.ping_timeout;
		ioctld->portals[p].replacement_timeout =
		    portal->session_timeout_options.replacement_timeout;
		ioctld->portals[p].disk_command_timeout =
		    portal->session_timeout_options.disk_command_timeout;
		ioctld->portals[p].abort_timeout =
		    portal->error_timeout_options.abort_timeout;
		ioctld->portals[p].reset_timeout =
		    portal->error_timeout_options.reset_timeout;

		ioctld->portals[p].initial_r2t =
		    portal->iscsi_options.InitialR2T;
		ioctld->portals[p].immediate_data =
		    portal->iscsi_options.ImmediateData;
		ioctld->portals[p].max_recv_data_segment_len =
		    portal->iscsi_options.MaxRecvDataSegmentLength;
		ioctld->portals[p].first_burst_len =
		    portal->iscsi_options.FirstBurstLength;
		ioctld->portals[p].max_burst_len =
		    portal->iscsi_options.MaxBurstLength;
		ioctld->portals[p].def_time2wait =
		    portal->iscsi_options.DefaultTime2Wait;
		ioctld->portals[p].def_time2retain =
		    portal->iscsi_options.DefaultTime2Retain;

		switch (portal->iscsi_options.HeaderDigest) {
		case CONFIG_DIGEST_NEVER:
			ioctld->portals[p].header_digest = ISCSI_DIGEST_NONE;
			break;
		case CONFIG_DIGEST_ALWAYS:
			ioctld->portals[p].header_digest = ISCSI_DIGEST_CRC32C;
			break;
		case CONFIG_DIGEST_PREFER_ON:
			ioctld->portals[p].header_digest =
			    ISCSI_DIGEST_CRC32C_NONE;
			break;
		case CONFIG_DIGEST_PREFER_OFF:
			ioctld->portals[p].header_digest =
			    ISCSI_DIGEST_NONE_CRC32C;
			break;
		}

		switch (portal->iscsi_options.DataDigest) {
		case CONFIG_DIGEST_NEVER:
			ioctld->portals[p].data_digest = ISCSI_DIGEST_NONE;
			break;
		case CONFIG_DIGEST_ALWAYS:
			ioctld->portals[p].data_digest = ISCSI_DIGEST_CRC32C;
			break;
		case CONFIG_DIGEST_PREFER_ON:
			ioctld->portals[p].data_digest =
			    ISCSI_DIGEST_CRC32C_NONE;
			break;
		case CONFIG_DIGEST_PREFER_OFF:
			ioctld->portals[p].data_digest =
			    ISCSI_DIGEST_NONE_CRC32C;
			break;
		}

		ioctld->portals[p].tcp_window_size =
		    portal->tcp_options.window_size;
		/*
		 * tmp hack - rest of daemon needs to be fixed for
		 * endien and ipv6 junk
		 */
		addr = (struct sockaddr_in *)&ioctld->portals[p].addr;
		addr->sin_family = AF_INET;
		memcpy(&addr->sin_addr.s_addr, portal->descriptor->ip,
		       portal->descriptor->ip_length);
		addr->sin_port = htons(portal->descriptor->port);
		ioctld->portals[p].tag = portal->descriptor->tag;
	}

	if (config->target->auth_options.username[0]
	    && config->target->auth_options.password_length) {
		ioctld->password_length =
		    config->target->auth_options.password_length;
		strncpy(ioctld->username, config->target->auth_options.username,
			sizeof (ioctld->username));
		ioctld->username[sizeof (ioctld->username) - 1] = '\0';
		memcpy(ioctld->password, config->target->auth_options.password,
		       MIN(config->target->auth_options.password_length,
			   sizeof (ioctld->password)));
		ioctld->password[sizeof (ioctld->password) - 1] = '\0';
		debugmsg(2,
			 "authentication outgoing username %s password %s "
			 "for target %s",
			 ioctld->username, ioctld->password,
			 config->target->TargetName);
		if (config->target->auth_options.username_in[0]
		    && config->target->auth_options.password_length_in) {
			/* also copy the target authentication information */
			ioctld->password_length_in =
			    config->target->auth_options.password_length_in;
			strncpy(ioctld->username_in,
				config->target->auth_options.username_in,
				sizeof (ioctld->username_in));
			ioctld->username_in[sizeof (ioctld->username_in) - 1] =
			    '\0';
			memcpy(ioctld->password_in,
			       config->target->auth_options.password_in,
			       MIN(config->target->auth_options.
				   password_length_in,
				   sizeof (ioctld->password_in)));
			ioctld->password_in[sizeof (ioctld->password_in) - 1] =
			    '\0';
			debugmsg(2,
				 "authentication incoming username %s "
				 "password %s for target %s",
				 ioctld->username_in, ioctld->password_in,
				 config->target->TargetName);
		}
	} else {
		ioctld->password_length = 0;
		ioctld->password_length_in = 0;
		debugmsg(2, "authentication not configured for target %s",
			 config->target->TargetName);
	}

	memcpy(ioctld->initiator_name, daemon_config.initiator_name,
	       strlen(daemon_config.initiator_name));
	memcpy(ioctld->initiator_alias, daemon_config.initiator_alias,
	       strlen(daemon_config.initiator_alias));
	memcpy(ioctld->target_name, config->target->TargetName,
	       strlen(config->target->TargetName));

	if (config->target->preferred_portal)
		memcpy(ioctld->preferred_portal,
		       config->target->preferred_portal,
		       strlen(config->target->preferred_portal));

	if (config->target->preferred_subnet) {
		memcpy(ioctld->preferred_subnet,
		       config->target->preferred_subnet,
		       strlen(config->target->preferred_subnet));
		ioctld->preferred_subnet_mask =
		    config->target->preferred_subnet_mask;
	}

	memcpy(ioctld->isid, config->isid, sizeof (ioctld->isid));
	ioctld->host_number = -1;
	ioctld->portal_failover = config->target->portal_failover;
	

	/*
	 * Tell the kernel to establish a session.
	 */
	do {
		rc = ioctl(control_fd, ISCSI_ESTABLISH_SESSION,
			   (char *) ioctld);
		if (rc < 0) {
			if (errno == EBUSY) {
				debugmsg(2,
					 "iSCSI session ioctl got EBUSY, "
					 "retrying");
				usleep(20 * 1000);
			} else if (errno == EINVAL) {
				ret = ISCSI_SESSION_FAILED_NO_RETRY;	
					/* most likely a version mismatch 
					 * with the kernel module 
					 */
				goto done;
			} else {
				errormsg
				    ("iSCSI session ioctl failed for %s, %s",
				     config->target->TargetName,
				     strerror(errno));
				sleep(1);
			}
		} else if (rc == 0) {
			/* inform the main daemon that the session failed, 
			 * but we don't want to fork another session process.  
			 */
			logmsg(AS_ERROR,
			       "iSCSI session ioctl failed for %s, rc 0",
			       config->target->TargetName);
			ret = ISCSI_SESSION_FAILED_NO_RETRY;
			goto done;
		} else
			debugmsg(2,
				 "session ioctl for %s returned from "
				 "kernel with rc %d",
				 config->target->TargetName, rc);

		if (stop_requested)
			goto done;

	} while (rc < 0);

	/* we only get the host and channel numbers if a new session claimed 
	 * them, so this tells us whether we updated an existing session or 
	 * started a new one.  If we started a new one, log it, and maybe 
	 * run a script.
	 */
	if (ioctld->host_number >= 0) {
		debugmsg(2, "host %d is target %s", ioctld->host_number,
			 config->target->TargetName);
		/* FIXME: call a target-available script/program? */
	}

      done:
	/* don't want to leave passwords in memory */
	memset(&ioctld, 0, sizeof (ioctld));
	free(ioctld);

	return ret;
}

/* handle any max target number limits in this function */
int
iscsi_session_process(struct iscsi_session_process *process,
		      struct iscsi_session_config *config)
{
	struct sigaction action;
	int ret;

	memset(&action, 0, sizeof (struct sigaction));
	action.sa_sigaction = NULL;
	action.sa_flags = 0;
	action.sa_handler = SIG_DFL;
	sigaction(SIGTERM, &action, NULL);
	sigaction(SIGINT, &action, NULL);

	action.sa_handler = SIG_IGN;
	sigaction(SIGCHLD, &action, NULL);

	ret = ioctl_establish_session(config, 1, process->config_number, 0);

	/* don't want to leave passwords in memory */
	memset(&config->target->auth_options, 0,
	       sizeof (config->target->auth_options));

	exit(ret);
}

void
ioctl_terminate_session(struct iscsi_session_config *config)
{
	debugmsg(1, "terminate session not implemented\n",
		 __FILE__, __LINE__);
}

void
iscsi_kill_session(struct iscsi_session_process *process)
{
	pid_t pid = process->pid;

	if (pid) {
		debugmsg(2, "sending SIGTERM to session process %p pid %d",
			 process, pid);
		/* if the proces is still running, kill it */
		kill(SIGTERM, pid);
	}

	/* if the daemon is dying, leave the session running.
	 * the sysfs shutdown will take care of it.
	 */
	if (stop_requested)
		return;
}


void
iscsi_lun_inventory_changed(struct iscsi_session_process *process)
{
	struct iscsi_session_config *config = process->config;
	pid_t pid = 0;

	if (config == NULL) {
		logmsg(AS_ERROR, "can't reprobe LUNs for process %p, no config",
		       process);
		return;
	}

      fork_again:
	pid = fork();
	if (pid == 0) {
		/* FIXME: ignore SIGHUP? */
		exit(0);
	} else if (pid < 0) {
		errormsg("cannot fork child");
		usleep(100 * 1000);
		goto fork_again;
	} else {
		logmsg(AS_NOTICE, "reprobing LUNs for bus %d target %d",
		       config->iscsi_bus, config->target_id);
	}
}

void
iscsi_config_changed(struct iscsi_session_process *process,
		     struct iscsi_session_config *new_config,
		     struct iscsi_session_config *current_config)
{
	int probe_luns = 0;
	pid_t pid;

	if (process == NULL) {
		logmsg(AS_ERROR,
		       "config changed for NULL process, new_config %p, "
		       "current_config %p\n",
		       new_config, current_config);
		return;
	}

	if (new_config == NULL) {
		/* This can happen only if target is deleted */
		ioctl_terminate_session(current_config);
		return;
	}

	logmsg(AS_NOTICE, "updating bus %d target %d to configuration #%lu",
	       new_config->iscsi_bus, new_config->target_id,
	       process->config_number);

	if ((new_config->target)
	    && (new_config->target->auth_options.password_length)) {
		debugmsg(2, "new config has username %s password %s length %d",
			 new_config->target->auth_options.username,
			 new_config->target->auth_options.password,
			 new_config->target->auth_options.password_length);
	}

	/* if a LUN probe is needed, we might as well do it now */
	if (process->lun_inventory_changed) {
		/* no point doing a separate ioctl */
		process->lun_inventory_changed = 0;
		probe_luns = 1;
		debugmsg(2,
			 "bus %d target %d LUN inventory may have changed, "
			 "reprobing LUNs while updating config",
			 new_config->iscsi_bus, new_config->target_id);
	}

	if ((new_config->iscsi_bus != current_config->iscsi_bus)
	    || (new_config->target_id != current_config->target_id)) {
		logmsg(AS_NOTICE,
		       "bus %d target %d rebound to bus %d target %d",
		       current_config->iscsi_bus, current_config->target_id,
		       new_config->iscsi_bus, new_config->target_id);
	}

      fork_again:
	pid = fork();
	if (pid == 0) {
		/* FIXME: ignore SIGHUP? */
		ioctl_establish_session(new_config, probe_luns,
					process->config_number, 1);

		if (new_config && new_config->target)
			memset(&new_config->target->auth_options, 0,
			       sizeof (new_config->target->auth_options));
		if (current_config && current_config->target)
			memset(&current_config->target->auth_options, 0,
			       sizeof (current_config->target->auth_options));
		debugmsg(2, "updated configuration for bus %d target %d",
			 new_config->iscsi_bus, new_config->target_id);
		exit(0);
	} else if (pid < 0) {
		errormsg("cannot fork child to update config");
		usleep(100 * 1000);
		goto fork_again;
	} else {
		/* give it a chance to get going */
		usleep(10 * 1000);
	}
}

void
iscsi_session_count_changed(int num_sessions, int discovery_in_progress,
			    int discovery_not_in_progress)
{
	/* we don't care, since sessions can take care of themselves */
}

int
iscsi_lock_file(int fd)
{
	struct flock lock;
	int result = 0;

	memset(&lock, 0, sizeof (lock));

	lock.l_type = F_WRLCK;

	/* entire file */
	lock.l_start = 0;
	lock.l_whence = SEEK_SET;
	lock.l_len = 0;

	/* try to acquire the lock, returning if we can't */
	result = fcntl(fd, F_SETLK, &lock);

	return result;
}

int
iscsi_unlock_file(int fd)
{
	struct flock lock;
	int result = 0;

	memset(&lock, 0, sizeof (lock));

	lock.l_type = F_UNLCK;

	/* entire file */
	lock.l_start = 0;
	lock.l_whence = SEEK_SET;
	lock.l_len = 0;

	/* try to release the lock */
	result = fcntl(fd, F_SETLKW, &lock);

	return result;
}

/* lock a pid file, to ensure that only one instance of the 
 * daemon can run at a time 
 */
int
lock_pid_file(void)
{
	int fd =
	    open(daemon_config.pid_file, O_CREAT | O_RDWR, S_IRUSR | S_IWUSR);
	char buffer[64];
	int val;

	if (fd < 0) {
		errormsg("open of lock_file %s failed", daemon_config.pid_file);
		exit(0);
	}

	debugmsg(7, "pid file fd %d", fd);

	if (iscsi_lock_file(fd) < 0) {
		if (errno == EACCES
		    || ((errno == EAGAIN) && !iscsi_process_should_exit())) {
			/* daemon already running */
			fprintf(stderr, "pid file %s already locked",
				daemon_config.pid_file);
			close(fd);
			exit(0);
		} else {
			/* lock error */
			errormsg("iscsi_lock_file %s failed",
				 daemon_config.pid_file);
			close(fd);
			exit(0);
		}
	} else {
		debugmsg(7, "locked pid file %s", daemon_config.pid_file);
	}

	/* truncate the file, and write our pid */
	ftruncate(fd, 0);
	sprintf(buffer, "%d\n", getpid());
	if (write(fd, buffer, strlen(buffer)) != strlen(buffer)) {
		errormsg("error writing pid file %s", daemon_config.pid_file);
		close(fd);
		exit(0);
	}

	/* sync the file to disk */
	fsync(fd);

	/* mark the pid file close-on-exec */
	if ((val = fcntl(fd, F_GETFD, 0)) < 0) {
		errormsg("failed to get pid file %s flags",
			 daemon_config.pid_file);
		close(fd);
		exit(0);
	}
	val |= FD_CLOEXEC;
	if (fcntl(fd, F_SETFD, val) < 0) {
		errormsg("error setting flags on pid file %s descriptor %d",
			 daemon_config.pid_file, fd);
		close(fd);
		exit(0);
	}

	return 1;
}

void
logmsg(int priority, const char *fmt, ...)
{
	va_list args;
	char buf[1024];
	int rc;

	va_start(args, fmt);
	rc = vsnprintf(buf, sizeof (buf), fmt, args);

	if ((rc > 0) && (rc < sizeof (buf))) {
		/* don't end with a double newline */
		if (buf[rc - 1] == '\n')
			buf[rc - 1] = '\0';
	}

	if (daemon_config.debug_level) {
		struct timeval time;

		gettimeofday(&time, NULL);
		fprintf(stderr, "%ld.%ld >> %s[%d]: %s\n", time.tv_sec,
			time.tv_usec, progname, getpid(), buf);
		fflush(stderr);
	}
	syslog(priority, "%s\n", buf);
	va_end(args);
}

void
errormsg(const char *fmt, ...)
{
	va_list args;
	char buf[1024];
	int en = errno;
	int rc;

	va_start(args, fmt);
	rc = vsnprintf(buf, sizeof (buf), fmt, args);

	if ((rc > 0) && (rc < sizeof (buf))) {
		/* don't end with a double newline */
		if (buf[rc - 1] == '\n')
			buf[rc - 1] = '\0';
	}

	if (daemon_config.debug_level) {
		struct timeval time;

		gettimeofday(&time, NULL);
		fprintf(stderr, "%ld.%ld >> %s[%d]: %s: %s\n", time.tv_sec,
			time.tv_usec, progname, getpid(), buf, strerror(en));
		fflush(stderr);
	}
	syslog(LOG_ERR, "%s: %s\n", buf, strerror(en));
	va_end(args);
}

void
debugmsg(int level, const char *fmt, ...)
{
	va_list args;
	char buf[1024];
	struct timeval time;
	int rc;

	if (daemon_config.debug_level < level)
		return;

	va_start(args, fmt);
	rc = vsnprintf(buf, sizeof (buf), fmt, args);

	if ((rc > 0) && (rc < sizeof (buf))) {
		/* don't end with a double newline */
		if (buf[rc - 1] == '\n')
			buf[rc - 1] = '\0';
	}

	gettimeofday(&time, NULL);
	fprintf(stderr, "%ld.%ld >> %s[%d]: %s\n", time.tv_sec, time.tv_usec,
		progname, getpid(), buf);
	fflush(stderr);

	/* syslog can block, and prevent the daemon from getting a new
	 * socket passed to the kernel in time, so we no longer syslog
	 * debug messages.  For other messages, we need to be careful to
	 * avoid having syslog block in the critical path.
	 */
	va_end(args);
}
