/*
 * iSCSI driver for Linux
 * Copyright (C) 2001 Cisco Systems, Inc.
 * Copyright (C) 2004 Red Hat, Inc.
 * maintained by linux-iscsi-devel@lists.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * iscsi-ls.c - User level utility to display iscsi devices.
 * $Id: iscsi-ls.c,v 1.1.2.5 2004/08/25 07:33:07 surekhap Exp $
 */

#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <errno.h>
#include <sys/time.h>
#include <sysfs/libsysfs.h>
#include "iscsi-version.h"
#include "iscsi-common.h"
#include "iscsi-ioctl.h"

#define TEXT_KEYS	1
#define LUN_INFO	2
#define HOST_SPECIFIC	4

#define SCSI_HOST_CLASS "scsi_host"

#define NIPQUAD(addr) \
	((unsigned char *)&addr)[0], \
	((unsigned char *)&addr)[1], \
	((unsigned char *)&addr)[2], \
 	((unsigned char *)&addr)[3]

/**
 * print_usage - Prints the usage of the iscsi-ls command incase of any error
 *
 */
void
print_usage(void)
{
	printf("\nUsage: iscsi-ls [-c] [-l] [-i host_id]\n");
	printf("-c : iSCSI target configuration parameters\n");
	printf("-i : Details regarding the specific host ID\n");
	printf("-l : Lun details available on the target\n");
	return;
}

/**
 * print_lun_info - prints the lun information for a target
 * @target_details: The target info structure for the current target
 *
 * The function prints the lun information for the target specified by the
 * target_details structure by reading from /proc/scsi/scsi file
 */
int
print_lun_info(struct target_info target_details)
{
	FILE *fp;
	char *rdptr, rdbuf[1024];
	int host_id, lun_id;

	fp = fopen("/proc/scsi/scsi", "r");
	if (!fp) {
		perror("\nCannot open /proc/scsi/scsi\n");
		return -1;
	}
	printf("\nDEVICE DETAILS :\n");
	printf("-------------- \n");
	do {
		rdptr = fgets(rdbuf, sizeof(rdbuf), fp);
		if (!rdptr || (*rdptr == '#'))
			continue;
		while (isspace(*rdptr)) {
			rdptr += 1;
		}
		if (!strncasecmp(rdptr, "Attached", 8))
			continue;
		if (!strncasecmp(rdptr, "Host: scsi", 10)) {
			sscanf(&rdptr[10], "%d", &host_id);
			if (host_id != target_details.host_no) {
				continue;
			}
			
			sscanf(&rdptr[36], "%d", &lun_id);
			printf("LUN ID : " "%d\n", lun_id);
			rdptr = fgets(rdbuf, sizeof(rdbuf), fp);
			printf("%s", rdptr);
			rdptr = fgets(rdbuf, sizeof(rdbuf), fp);
			printf("%s", rdptr);
		}
	} while (rdptr);
	fclose(fp);
	return 0;
}

/**
 * print_portal_info - Print the iscsi portal related info
 * @ctlfd: The open file descriptor for calling the ioctl function
 * @target_details: Details of the target for which portal info is displayed
 *
 * The ip address, port and the portal tag for the target is printed
 */
int
print_portal_info(int ctlfd, struct target_info target_details)
{
	struct sockaddr_in *addr;
	struct iscsi_portal_list portal_list;
	int j;

	printf("NO. OF PORTALS          : %d \n", target_details.num_portals);
	memset(&portal_list, 0, sizeof(portal_list));

	portal_list.host_no = target_details.host_no;
	if (ioctl(ctlfd, ISCSI_LS_PORTAL_INFO, &portal_list)) {
		fprintf(stderr, "The iscsi driver is inoperative\n");
		free(portal_list.portals);
		return -1;
	}
	for (j = 0; j < target_details.num_portals; j++) {
		addr = (struct sockaddr_in *)&portal_list.portals[j].addr;

		printf("PORTAL ADDRESS %d        : %u.%u.%u.%u:%d,%d\n",
		       (j + 1), NIPQUAD(addr->sin_addr.s_addr),
		       ntohs(addr->sin_port), portal_list.portals[j].tag);
	}

	return 0;
}

/**
 * print_common_info - Prints the name and connection/session status info
 * @target_details: Details of the target for which the info is displayed
 * @ctlfd: The open file descriptor for ioctl
 *
 * This function prints the information related to the target name, alias,
 * host number, target address, session status, session id
 */
int
print_common_info(int ctlfd, struct target_info target_details)
{
	struct timeval tv;
	struct sockaddr_in *addr;
	addr = (struct sockaddr_in *)&target_details.session_data.addr;

	printf("TARGET NAME             : %s\n", target_details.target_name);
	printf("TARGET ALIAS            : %s\n", target_details.target_alias);
	printf("HOST NO                 : %d \n", target_details.host_no);
	printf("TARGET ADDRESS          : %u.%u.%u.%u:%d\n",
	       NIPQUAD(addr->sin_addr.s_addr), ntohs(addr->sin_port));
	if (!target_details.session_data.session_alive) {
		if (!target_details.session_data.ever_established) {
			printf("SESSION STATUS          : CONNECTION ON "
			       "%u.%u.%u.%u:%d\n                          "
			       "NOT ESTABLISHED\n",
                               NIPQUAD(addr->sin_addr.s_addr),
                               ntohs(addr->sin_port));
		} else {
			gettimeofday(&tv, NULL);
			tv.tv_sec -= target_details.session_data.
				     session_drop_time;
			printf("SESSION STATUS          : CONNECTION ON "
			       "%u.%u.%u.%u:%d\n                          "
			       "DROPPED AT %s",
                               NIPQUAD(addr->sin_addr.s_addr),
                               ntohs(addr->sin_port),
			       ctime((time_t *) & (tv.tv_sec)));
		}
	} else {
		gettimeofday(&tv, NULL);
		tv.tv_sec -= target_details.session_data.establishment_time;
		printf("SESSION STATUS          : CONNECTION ON %u.%u.%u.%u:%d"
			"\n                          ESTABLISHED AT %s",
		       NIPQUAD(addr->sin_addr.s_addr),
                       ntohs(addr->sin_port),	
		       ctime((time_t *) & (tv.tv_sec)));
	}
	if (print_portal_info(ctlfd, target_details) < 0)
		return -1;
	printf("SESSION ID              : ISID %02x%02x%02x%02x%02x%02x TSID "
	       "%02x\n", target_details.session_data.isid[0],
	       target_details.session_data.isid[1],
	       target_details.session_data.isid[2],
	       target_details.session_data.isid[3],
	       target_details.session_data.isid[4],
	       target_details.session_data.isid[5],
	       target_details.session_data.tsid);
	return 0;
}

/**
 * print_config_info - Print information when the -c option is used
 * @target_details: Details of the target for which the config info is displayed
 *
 * The function prints the session keys that are configured for the particular
 * target specified by 'target_details'
 */
void
print_config_info(struct target_info target_details)
{
	printf("\nTARGET CONFIGURATION KEYS :\n");
	printf("------------------------- \n");
	printf("INITIAL R2T             : %s\n",
	       (target_details.session_data.initial_r2t) ? "Yes" : "No");
	printf("IMMEDIATE DATA          : %s\n",
	       (target_details.session_data.immediate_data) ? "Yes" : "No");
	printf("HEADERDIGEST            : %s\n",
	       (target_details.session_data.header_digest) ? "CRC32" : "NONE");
	printf("DATA DIGEST             : %s\n",
	       (target_details.session_data.data_digest) ? "CRC32" : "NONE");
	printf("FIRSTBURSTLENGTH        : %d\n",
	       target_details.session_data.first_burst_len);
	printf("MAXBURSTLENGTH          : %d\n",
	       target_details.session_data.max_burst_len);
	printf("MAXRECVDATASEGMENTLENGTH: initiator - %d\n",
	       target_details.session_data.max_recv_data_segment_len);
	printf("                          target    - %d\n",
	       target_details.session_data.max_xmit_data_segment_len);
	printf("LOGIN TIMEOUT           : %d\n",
	       target_details.session_data.login_timeout);
	printf("AUTH TIMEOUT            : %d\n",
	       target_details.session_data.auth_timeout);
	printf("ACTIVE TIMEOUT          : %d\n",
	       target_details.session_data.active_timeout);
	printf("IDLE TIMEOUT            : %d\n",
	       target_details.session_data.idle_timeout);
	printf("PING TIMEOUT            : %d\n",
	       target_details.session_data.ping_timeout);
}

 int
 find_host_ids(int **host_ids, int *host_count)
 {
	 struct sysfs_class *sfclass;
	 struct dlist *dev_list;
	 struct sysfs_class_device *cldev;
	 struct sysfs_attribute *attr;
	 int count = 0, i;
	 int *host_array = NULL;
	 char garbage[5];
	 
	 if(!(sfclass = sysfs_open_class(SCSI_HOST_CLASS))) {
		 fprintf(stderr, "Error opening scsi_host class in sysfs\n");
		 return -1;
	 }
	 if(!(dev_list = sysfs_get_class_devices(sfclass))) {
		 fprintf(stderr, "Error opening scsi_host class's device list "
			 "in sysfs\n");
		 return -1;
	 }
	 
	 count = dev_list->count;
	 
	 if(!(host_array = malloc(sizeof(int) * count))) {
		 fprintf(stderr, "Out of memory allocating host array");
		 return -1;
	 }
	 i = 0;
	 dlist_for_each_data(dev_list, cldev, struct sysfs_class_device) {
		 attr = sysfs_get_classdev_attr(cldev, "proc_name");
		 if (!strncmp(attr->value, "iscsi-sfnet", 11)) {
			 sscanf(cldev->name, "%4s%d", garbage, &(host_array[i])
				);
			 i++;
		 }
	 }
	 sysfs_close_class(sfclass);
	 *host_ids = host_array;
	 *host_count = i;
	 return 0;
 }

/**
 * display_info - Function to display all info based on the options
 * @options: The command line options given for iscsi-ls command
 * @ctlfd: The open file descriptor for the ioctl fuction
 *
 * This function gathers information for the particular target into target
 * details and based on the options specified for the command, the appropriate
 * information is displayed
 */
int
display_info(int options, int ctlfd, int host)
{
	struct target_info target_details;
	int *host_id, host_count, i;

	if (options & HOST_SPECIFIC) {
		host_id = (int *) malloc(sizeof(int));
		host_id[0] = host;
		host_count = 1;
	}
	else if(find_host_ids(&host_id, &host_count) < 0)
		return -1;
        
	for (i=0; i<host_count; i++) {
		memset(&target_details, 0, sizeof(target_details));
		target_details.host_no = host_id[i];
		
		if (ioctl(ctlfd, ISCSI_LS_TARGET_INFO, &target_details)) {
			if(!(options & HOST_SPECIFIC) && (errno == EINVAL)) {
				/* The host ID was not valid */
				continue;
			} else {
				fprintf(stderr,
					"The iscsi driver is inoperative\n");
				if(host_id)
					free(host_id);
				return -1;
			}
		}
		
		if (print_common_info(ctlfd, target_details) < 0)
			return -1;
		if (options & TEXT_KEYS) {
			print_config_info(target_details);
		}
		if (options & LUN_INFO) {
			print_lun_info(target_details);
		}
		printf("***********************************************"
		       "********************************\n");
	}
	if (host_id)
		free(host_id);

	return 0;
}
	
int
main(int argc, char **argv)
{
	int c, options = 0;
	int ctlfd;
	int host_id = -1;
	char version[256];

	while ((c = getopt(argc, argv, "lci:")) >= 0) {
		switch (c) {
		case 'c':
			options |= TEXT_KEYS;
			break;
		case 'i':
			options |= HOST_SPECIFIC;
			if(host_id >= 0) {
				fprintf(stderr, "-i option can only be"
					" specified once\n");
				print_usage();
				return -1;
			}
			host_id = atoi(optarg);
                       break;
		case 'l':
			options |= LUN_INFO;
			break;
		default:
			print_usage();
			return -1;
		}
	}
	ctlfd = open("/dev/iscsictl", O_RDONLY);
	if (ctlfd < 0) {
		printf
		    ("\n\n#######################################"
		     "########################################\n");
		printf("\niSCSI driver is not loaded\n");
		printf
		    ("\n\n#######################################"
		     "########################################\n\n\n");
		return -1;
	}

	sprintf(version, "SFNet iSCSI Driver Version ... %s (%s)",
		ISCSI_DRIVER_VERSION, ISCSI_DATE);

	printf
	    ("**************************************"
	     "*****************************************\n");
	printf("\t%s\n", version);
	printf
	    ("**************************************"
	     "*****************************************\n");
	if (display_info(options, ctlfd, host_id) < 0)
		return -1;
	close(ctlfd);
	return 0;
}
